# This file makes Python treat the directory as a package.

# Import contents of the PyIndi module into the package namespace
# for backward compatibility (allows 'import PyIndi')
from .PyIndi import *
